/**
 * CMI : Cluster Method Invocation
 * Copyright (C) 2007,2008 Bull S.A.S.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: AbsServerClusterViewManager.java 1796 2008-05-16 16:05:05Z eyindanga $
 * --------------------------------------------------------------------------
 */

package org.ow2.cmi.controller.server;

import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.rmi.NoSuchObjectException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import javax.ejb.EJBObject;
import javax.management.remote.JMXServiceURL;
import javax.naming.Context;
import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;

import net.jcip.annotations.ThreadSafe;

import org.ow2.cmi.admin.CMIAdminConnectorManager;
import org.ow2.cmi.admin.CMIMBeanConfigException;
import org.ow2.cmi.admin.MBeanUtils;
import org.ow2.cmi.config.CMIConfig;
import org.ow2.cmi.controller.common.AbsClusterViewManager;
import org.ow2.cmi.controller.common.ClusterViewManager;
import org.ow2.cmi.controller.provider.ClientClusterViewProvider;
import org.ow2.cmi.controller.provider.ClusteredClientClusterViewProvider;
import org.ow2.cmi.controller.server.manager.DefaultPolicyStrategyManager;
import org.ow2.cmi.info.CMIInfoExtractor;
import org.ow2.cmi.info.CMIInfoExtractorException;
import org.ow2.cmi.info.CMIInfoRepository;
import org.ow2.cmi.info.ClusteredObjectInfo;
import org.ow2.cmi.lb.data.PolicyData;
import org.ow2.cmi.lb.policy.FirstAvailable;
import org.ow2.cmi.lb.policy.HASingleton;
import org.ow2.cmi.lb.policy.IPolicy;
import org.ow2.cmi.lb.policy.Random;
import org.ow2.cmi.lb.policy.RoundRobin;
import org.ow2.cmi.lb.strategy.IStrategy;
import org.ow2.cmi.lb.strategy.LoadFactorComparator;
import org.ow2.cmi.lb.strategy.LoadFactorSort;
import org.ow2.cmi.lb.strategy.LocalPreference;
import org.ow2.cmi.lb.strategy.NoStrategy;
import org.ow2.cmi.lb.util.PolicyFactory;
import org.ow2.cmi.loader.DefaultPolicyStrategyLoader;
import org.ow2.cmi.loader.IPolicyStrategyLoader;
import org.ow2.cmi.reference.CMIReference;
import org.ow2.cmi.reference.ObjectNotFoundException;
import org.ow2.cmi.reference.ServerId;
import org.ow2.cmi.reference.ServerRef;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.pool.api.IPoolConfiguration;

/**
 * Common implementation for all manager of the cluster view at server-side.
 *
 * @author The new CMI team
 */
@ThreadSafe
public abstract class AbsServerClusterViewManager extends AbsClusterViewManager
implements ServerClusterViewManager {

    /**
     * Logger.
     */
    private static final Log LOGGER = LogFactory
    .getLog(AbsServerClusterViewManager.class);

    /**
     * Map names of protocol and InitialContextFactories in purpose to perform
     * lookups.
     */
    private final Map<String, String> initialContextFactories = new ConcurrentHashMap<String, String>();

    /**
     * References on the local registry for each protocol.
     */
    private final Map<String, ServerId> refsOnLocalRegistries = new ConcurrentHashMap<String, ServerId>();

    /**
     * References on the cluster view provider for each protocol.
     */
    private final Map<String, Remote> clientClusterViewProviders = new ConcurrentHashMap<String, Remote>();

    /**
     * References on the context for each protocol to perform bindings or
     * unbindings.
     */
    private final Map<String, Context> contexts = new ConcurrentHashMap<String, Context>();

    /**
     * Address of the local registry.
     */
    private InetAddress inetAddress = null;

    /**
     * Informations to replicate the dummy registry.
     */
    private ClusteredObjectInfo clusteredObjectInfoForDummyRegistry;

    /**
     * True if the replication manager is started.
     */
    private boolean replicationManagerStarted = false;

    /**
     * True if the server-side manager is started.
     */
    private static boolean started = false;

    /**
     * Store each interfaces.
     */
    private final ConcurrentHashMap<String, Class<?>> interfaces = new ConcurrentHashMap<String, Class<?>>();

    /**
     * Manager for Load balancing policies and strategies.
     */
    protected IPolicyStrategyManager policyStrategyManager = null;

    /**
     * Loader for Load balancing policies and strategies.
     */
    protected IPolicyStrategyLoader policyStrategyLoader = null;

    /**
     * Store each class of Policy.
     */
    @SuppressWarnings("unchecked")
    private final ConcurrentHashMap<String, Class<? extends IPolicy>> policyMap =
        new ConcurrentHashMap<String, Class<? extends IPolicy>>();

    /**
     * Store each class of Strategy.
     */
    @SuppressWarnings("unchecked")
    private final ConcurrentHashMap<String, Class<? extends IStrategy>> strategyMap =
        new ConcurrentHashMap<String, Class<? extends IStrategy>>();

    /**
     * Embedded Policy classes.
     */
    private final Class<?>[] embeddedPolicyClasses = {FirstAvailable.class,
            HASingleton.class, Random.class, RoundRobin.class };

    /**
     * Embedded Strategy classes.
     */
    private final Class<?>[] embeddedStrategyClasses = {NoStrategy.class,
            LoadFactorSort.class, LoadFactorComparator.class,
            LocalPreference.class };

    /**
     * Create and start the server-side manager.
     *
     * @return the cluster view manager
     * @throws ServerClusterViewManagerException
     *             if the instance of ServerClusterViewManager cannot be get
     */
    public static synchronized ServerClusterViewManager start()
    throws ServerClusterViewManagerException {
        ClusterViewManager clusterViewManager = AbsClusterViewManager
        .getClusterViewManager();
        if (!started && clusterViewManager != null) {
            LOGGER
            .warn("CMI has been already started with the creation of a new initial context.");
            started = true;
            if (!(clusterViewManager instanceof ServerClusterViewManager)) {
                throw new ServerClusterViewManagerException(
                        "A cluster view manager already exists but doesn't implement the interface ServerClusterViewManager");
            }
        } else if (!started) {
            clusterViewManager = getServerClusterViewManager();
            ((AbsServerClusterViewManager) clusterViewManager).doStart();
            started = true;
        }
        return (ServerClusterViewManager) clusterViewManager;
    }

    /**
     * Stop the server-side manager.
     */
    public synchronized void stop() {
        if (started) {
            AbsServerClusterViewManager serverClusterViewManager = (AbsServerClusterViewManager) getServerClusterViewManager();
            // Unregister the CMI MBean
            try {
                MBeanUtils.unregisterCMIMBean();
            } catch (CMIMBeanConfigException e) {
                LOGGER.error("Cannot unregister CMIMBean", e);
            }
            // Unbind and unexport cmi objects
            for (String protocol : serverClusterViewManager.initialContextFactories
                    .keySet()) {
                try {
                    serverClusterViewManager.contexts.get(protocol).unbind(
                            CMIConfig.getBindNameForProvider());
                } catch (NamingException e) {
                    LOGGER
                    .error(
                            "Cannot unbind the instance of ClientClusterViewProvider for the protocol {0}",
                            protocol, e);
                }
                // Unexport the connector for only this protocol
                // So disable the MultiPRODelegate provided by Carol
                System.setProperty("carol.multipro.protocol", protocol);
                try {
                    PortableRemoteObject
                    .unexportObject(serverClusterViewManager.clientClusterViewProviders
                            .get(protocol));
                } catch (NoSuchObjectException e) {
                    LOGGER
                    .error(
                            "Cannot unexport the instance of ClientClusterViewProvider for the protocol {0}",
                            protocol, e);
                } finally {
                    System.setProperty("carol.multipro.protocol", "any");
                }
            }
            serverClusterViewManager.doStop();
            AbsClusterViewManager.setClusterViewManager(null);
            started = false;
            LOGGER.info("The server-side manager is now stopped.");
        }
    }

    /**
     * Start the server-side manager. This method is overridden by the
     * implementation of the server.
     */
    public abstract void doStart();

    /**
     * Stop the server-side manager. This method is overridden by the
     * implementation of the server.
     */
    public abstract void doStop();

    /**
     * @return a ServerClusterViewManager instance.
     * @throws ServerClusterViewManagerException
     *             if the instance of ServerClusterViewManager cannot be get.
     */
    public static final synchronized ServerClusterViewManager getServerClusterViewManager()
    throws ServerClusterViewManagerException {
        // Checks if an instance already exists
        ClusterViewManager clusterViewManager = AbsClusterViewManager
        .getClusterViewManager();
        if (clusterViewManager == null) {

            // Initializes CMIMBean
            try {
                MBeanUtils.initCMIMBean();
            } catch (CMIMBeanConfigException e) {
                LOGGER.error("Cannot initializes CMIMBean", e);
            }

            // Loads the implementation of ServerClusterViewManager via
            // CMIConfig
            Class<? extends ServerClusterViewManager> serverClusterViewManagerClass;
            try {
                serverClusterViewManagerClass = CMIConfig
                .getServerClusterViewManagerClass();
            } catch (ClassNotFoundException e) {
                LOGGER
                .error("Cannot load the class implementing a server-side manager.");
                throw new ServerClusterViewManagerException(
                        "Cannot load the class implementing a a server-side manager.",
                        e);
            }
            LOGGER.debug("The ServerClusterViewManager is: {0}",
                    serverClusterViewManagerClass.getName());

            AbsServerClusterViewManager serverClusterViewManager;
            try {
                serverClusterViewManager = (AbsServerClusterViewManager) serverClusterViewManagerClass
                .newInstance();
            } catch (Exception e) {
                LOGGER
                .error(
                        "Cannot create the instance (singleton) of ServerClusterViewManager",
                        e);
                throw new ServerClusterViewManagerException(
                        "Cannot get an instance of ServerClusterViewManager", e);
            }

            LOGGER.debug("ServerClusterViewManager has been created");

            Class<? extends IPolicyStrategyManager> policyStrategyManager = null;
            try {
                policyStrategyManager = CMIConfig
                .getPolicyStrategyManagerClass();
            } catch (ClassNotFoundException e) {
                LOGGER
                .error("Cannot load the class implementing a policy/strategy manager.");
                throw new ServerClusterViewManagerException(
                        "Cannot load the class implementing a policy/strategy manager.",
                        e);
            }
            LOGGER.debug("The Manager for policies and strategies is: {0}",
                    policyStrategyManager.getName());
            IPolicyStrategyManager policStrategyManager = null;
            try {
                if (policyStrategyManager.getName().equals(
                        CMIConfig.DEFAULT_POLICY_STRATEGY_MANAGER)) {
                    policStrategyManager = policyStrategyManager.newInstance();
                    serverClusterViewManager
                    .setPolicyStrategyManager(policStrategyManager);
                    ((DefaultPolicyStrategyManager) policStrategyManager)
                    .setServerClusterViewManager(serverClusterViewManager);
                }
            } catch (Exception e) {
                LOGGER
                .error("Cannot instantiate default class implementing a policy/strategy manager.");
                throw new ServerClusterViewManagerException(
                        "Cannot instantiate default class implementing a policy/strategy manager.",
                        e);
            }
            Class<? extends IPolicyStrategyLoader> policyStrategyLoader = null;
            try {
                policyStrategyLoader = CMIConfig.getPolicyStrategyLoaderClass();
            } catch (ClassNotFoundException e) {
                LOGGER
                .error("Cannot load the class implementing a policy/strategy loader.");
                throw new ServerClusterViewManagerException(
                        "Cannot load the class implementing a policy/strategy loader.",
                        e);
            }
            LOGGER.debug("The Loader for policies and stretgies is: {0}",
                    policyStrategyLoader.getName());
            IPolicyStrategyLoader pStrategyLoader = null;
            try {
                if (policyStrategyLoader.getName().equals(
                        CMIConfig.DEFAULT_POLICY_STRATEGY_LOADER)) {
                    pStrategyLoader = policyStrategyLoader.newInstance();
                    serverClusterViewManager
                    .setPolicyStrategyLoader(pStrategyLoader);
                    /**
                     * The policy loader will communicate with policy/strategy
                     * manager.
                     */
                    ((DefaultPolicyStrategyLoader) pStrategyLoader)
                    .setPolicyStrategyManager(policStrategyManager);
                }
            } catch (Exception e) {
                LOGGER
                .error("Cannot instantiate default class implementing a policy/strategy loader.");
                throw new ServerClusterViewManagerException(
                        "Cannot instantiate default class implementing a policy/strategy loader.",
                        e);
            }

            serverClusterViewManager.loadEmbeddedLBClasses();

            // Extracts informations on clustering of dummy context
            try {
                serverClusterViewManager.clusteredObjectInfoForDummyRegistry = CMIInfoExtractor
                .extractClusteringInfoFromAnnotatedPOJO(CMIConfig
                        .getBindNameForDummyRegistry(), null,
                        ClusteredDummyRegistry.class, false, false,
                        null);
            } catch (CMIInfoExtractorException e) {
                LOGGER.error("Cannot get infos for dummy context", e);
                throw new ServerClusterViewManagerException(
                        "Cannot get infos for dummy context", e);
            }
            // Shares delay to refresh
            serverClusterViewManager.setDelayToRefresh(CMIConfig
                    .getRefreshTime());
            // Registers CMIMBean
            try {
                MBeanUtils.registerCMIMBean(serverClusterViewManager);
            } catch (CMIMBeanConfigException e) {
                LOGGER.error("Cannot register CMIMBean", e);
            }

            serverClusterViewManager.initStats();

            clusterViewManager = serverClusterViewManager;
            AbsClusterViewManager.setClusterViewManager(clusterViewManager);

            LOGGER.info("The server-side manager was successfully started");
        } else {
            if (!(clusterViewManager instanceof ServerClusterViewManager)) {
                LOGGER
                .error("An instance of the manager that is not a ServerClusterViewManager already exists in the JVM");
                throw new ServerClusterViewManagerException(
                "An instance of the manager that is not a ServerClusterViewManager already exists in the JVM");
            }
        }
        return (ServerClusterViewManager) clusterViewManager;
    }

    /**
     * Adds a new protocol. This method also binds a provider of the cluster
     * view for clients and a dummy context, into the given instance of
     * CMIContext.
     *
     * @param initialContextFactoryName
     *            a InitialContextFactory for this protocol
     * @param serverRef
     *            a reference onto the local registry
     * @param cmiContext
     *            a instance of CMIContext to bind a provider of the cluster
     *            view for clients and a dummy context
     * @throws ServerClusterViewManagerException
     *             if the protocol cannot be added
     */
    public final void addProtocol(final String initialContextFactoryName,
            final ServerId serverRef, final Context cmiContext)
    throws ServerClusterViewManagerException {

        String protocol = serverRef.getProtocol();
        String providerURL = serverRef.getProviderURL();

        synchronized (initialContextFactoryName) {
            // Checks if the protocol has been already added
            if (!initialContextFactories.containsKey(protocol)) {
                LOGGER.debug("New protocol {0} added", protocol);
                initialContextFactories
                .put(protocol, initialContextFactoryName);

                // Binds a provider of the cluster view depending on the
                // configuration
                if (CMIConfig.isProviderBound()) {
                    LOGGER
                    .debug(
                            "Binds a provider of the cluster view for protocol {0}",
                            protocol);
                    bindClientClusterViewProvider(cmiContext, protocol);
                }

                // Bind a dummy registry to enable the load-balancing of JNDI
                // accesses for unclustered objects
                if (CMIConfig.isRegistryBound()) {
                    LOGGER
                    .debug(
                            "Binds a dummy object that represents the registry for protocol {0}",
                            protocol);
                    bindDummyRegistry(serverRef);
                }

                // Keeps a reference on the local registry (for local
                // preference)
                refsOnLocalRegistries.put(protocol, serverRef);
                InetAddress newInetAddress = serverRef.getInetAddress();
                try {
                    if (NetworkInterface.getByInetAddress(newInetAddress) == null) {
                        LOGGER.error("The referenced server is not local");
                        throw new ServerClusterViewManagerException(
                        "The referenced server is not local");
                    }
                } catch (SocketException e) {
                    LOGGER.error("Cannot know if the IP is local", e);
                    throw new ServerClusterViewManagerException(
                            "Cannot know if the IP is local", e);
                }
                if (inetAddress == null) {
                    LOGGER
                    .debug("InetAdress of manager is {0}",
                            newInetAddress);
                    inetAddress = newInetAddress;
                } else if (!newInetAddress.equals(inetAddress)) {
                    LOGGER
                    .error(
                            "Host name expected: {0} - Host name found : {1}. All the protocol have to use the same host name !",
                            inetAddress, newInetAddress);
                    throw new ServerClusterViewManagerException(
                            "Host name expected: "
                            + inetAddress.getHostName()
                            + " - Host name found : "
                            + newInetAddress.getHostName()
                            + ". All the protocol have to use the same host name !");
                }

                // Set load factor
                setLoadFactor(serverRef, CMIConfig.getLoadFactor());

                // Bind a JMX connector server for this protocol, if CMI is not
                // embedded and if the property
                if (!CMIConfig.isEmbedded()
                        && CMIConfig.isConnectorEnabled(protocol)) {
                    try {
                        CMIAdminConnectorManager.startConnector(protocol,
                                providerURL);
                    } catch (CMIMBeanConfigException e) {
                        LOGGER
                        .error(
                                "The connector for the protocol {0} cannot be added",
                                protocol, e);
                        throw new ServerClusterViewManagerException(
                                "The connector for the protocol " + protocol
                                + " cannot be added", e);
                    }
                }
            }
        }
    }

    /**
     * Binds an instance of {@link ClientClusterViewProvider} into a CMIContext
     * to enable replication of the provider.
     *
     * @param cmiContext
     *            a context that enables replication of the provider
     * @param protocol
     *            the protocol name
     * @throws ServerClusterViewManagerException
     *             if the provider cannot be bound
     * @see ClientClusterViewProvider
     */
    private void bindClientClusterViewProvider(final Context cmiContext,
            final String protocol) throws ServerClusterViewManagerException {

        ClientClusterViewProvider clientClusterViewProvider = null;

        // Export the connector for only this protocol
        // So disable the MultiPRODelegate provided by Carol
        System.setProperty("carol.multipro.protocol", protocol);
        try {
            clientClusterViewProvider = new ClusteredClientClusterViewProvider(
                    this);
        } catch (RemoteException e) {
            LOGGER
            .error(
                    "Cannot export the instance of ClientClusterViewProvider for the protocol {0}",
                    protocol, e);
            throw new ServerClusterViewManagerException(
                    "Cannot export the instance of ClientClusterViewProvider for the protocol "
                    + protocol, e);
        } finally {
            System.setProperty("carol.multipro.protocol", "any");
        }

        String bindName = CMIConfig.getBindNameForProvider();

        // Extracts data on load-balancing from the annotated class
        ClusteredObjectInfo clusteredObjectInfo;
        try {
            clusteredObjectInfo = CMIInfoExtractor
            .extractClusteringInfoFromAnnotatedPOJO(bindName,
                    ClientClusterViewProvider.class,
                    ClusteredClientClusterViewProvider.class, false,
                    false, null);
        } catch (CMIInfoExtractorException e) {
            LOGGER.error("Cannot get infos for client provider", e);
            throw new ServerClusterViewManagerException(
                    "Cannot get infos for client provider", e);
        }

        // Registers the data into the repository to enable replication
        CMIInfoRepository.addClusteredObjectInfo(bindName, clusteredObjectInfo);

        // Performs the binding into the CMIContext
        try {
            cmiContext.rebind(bindName, clientClusterViewProvider);
        } catch (NamingException e) {
            LOGGER
            .error(
                    "Cannot rebind a ClientClusterViewProvider for protocol {0}",
                    protocol, e);
            throw new ServerClusterViewManagerException(
                    "Cannot rebind a ClientClusterViewProvider " + protocol, e);
        }
        // Keep a reference on this pro in order to unexport it
        clientClusterViewProviders.put(protocol, clientClusterViewProvider);
        // Keep a reference on this pro in order to unbind it
        contexts.put(protocol, cmiContext);

        LOGGER
        .debug(
                "ClientClusterViewProvider has been rebound for the protocol {0} with the name {1}",
                protocol, bindName);
    }

    /**
     * Binds a dummy registry to enable the load-balancing of JNDI accesses to
     * unclustered objects.
     *
     * @param serverRef
     *            a reference on the local registry
     */
    private void bindDummyRegistry(final ServerId serverRef) {

        String bindName = CMIConfig.getBindNameForDummyRegistry();
        // Create an instance of CMIReference for this object
        CMIReference cmiReference = new CMIReference(serverRef, bindName);
        addObjectInstance(clusteredObjectInfoForDummyRegistry, cmiReference);
    }

    /**
     * Adds an object with the given name to a set of objects that need to have
     * their state up-to-date.
     *
     * @param objectName
     *            a name of object
     * @throws ObjectNotFoundException
     *             if none object has the given name
     */
    public final synchronized void addObjectToWatch(final String objectName)
    throws ObjectNotFoundException {
        if (!isWatched(objectName)) {
            LOGGER.debug("Adding {0} to the set of watched object", objectName);
            watch(objectName);
            updatePolicy(objectName);
        }
    }

    /**
     * Adds an instance of a clustered object. If an other instance already
     * exists, the current configuration is used.
     *
     * @param clusteredObjectInfo
     *            informations on a clustered object
     * @param cmiReference
     *            a reference on a clustered object
     */
    @SuppressWarnings("unchecked")
    public final synchronized void addObjectInstance(
            final ClusteredObjectInfo clusteredObjectInfo,
            final CMIReference cmiReference) {

        // Retrieves the informations for the given clustered object
        Class<?> itfClass = clusteredObjectInfo.getItfClass();
        String businessName = null;
        String itfName = null;

        if (itfClass != null) {
            Class<? extends EJBObject> businessClass = clusteredObjectInfo
            .getBusinessClass();
            itfName = itfClass.getName();

            // Keep the class
            interfaces.putIfAbsent(itfName, itfClass);
            if (businessClass != null) {
                businessName = businessClass.getName();
                interfaces.putIfAbsent(businessName, businessClass);
            }
        }

        Class<? extends IPolicy> policyClass = clusteredObjectInfo
        .getPolicyType();
        Class<? extends IStrategy> strategyClass = clusteredObjectInfo
        .getStrategyType();

        String clusterName = clusteredObjectInfo.getClusterName();
        IPoolConfiguration poolConfiguration = clusteredObjectInfo
        .getPoolConfiguration();
        String policyType = policyClass.getName();
        String strategyType = strategyClass.getName();
        Map<String, Object> properties = clusteredObjectInfo.getProperties();
        boolean hasState = clusteredObjectInfo.hasState();
        boolean replicated = clusteredObjectInfo.isReplicated();
        Set<String> applicationExceptionNames = clusteredObjectInfo
        .getApplicationExceptionNames();

        // Check correctness of policy by creating it
        new PolicyFactory<CMIReference>(this).getPolicy(policyClass,
                strategyClass, properties);

        // Encapsulates data on load-balancing
        PolicyData policyData = new PolicyData(policyType, strategyType,
                properties);

        // Encapsulates data to replicate
        String objectName = cmiReference.getObjectName();

        DistributedObjectInfo distributedObjectInfo = new DistributedObjectInfo(
                clusterName, objectName, itfName, businessName, policyData,
                hasState, replicated, applicationExceptionNames,
                poolConfiguration);
        /**
         * keep default policy and strategy for the given object.
         */
        addDefaultPolicyStrategy(objectName, policyType, strategyType);
        /*
         * Sets the informations that will be replicated.
         */
        addDistributedObjectInfo(objectName, distributedObjectInfo);

        // Adds the new reference
        LOGGER.debug("Adding {0}...", cmiReference);
        addCMIReference(cmiReference);
    }

    /**
     * add default policy and strategy for the <code>objectName</code>
     * parameter.
     *
     * @param objectName
     *            name of the object.
     * @param policyType
     *            the policy name
     * @param strategyType
     *            the strategy name
     */
    private void addDefaultPolicyStrategy(final String objectName,
            final String policyType, final String strategyType) {
        policyStrategyManager.addDefaultPolicyClassName(objectName, policyType);
        policyStrategyManager.addDefaultStrategyClassName(objectName,
                strategyType);
    }

    /**
     * Returns true if the given object is already replicated.
     *
     * @param objectName
     *            a name of object
     * @return true if the given object is already replicated
     */
    protected abstract boolean containObject(String objectName);

    /**
     * Returns informations on this object.
     *
     * @param objectName
     *            a name of object
     * @return informations on this object
     * @throws ObjectNotFoundException
     *             if the given object is not found
     */
    protected abstract DistributedObjectInfo getDistributedObjectInfo(
            String objectName) throws ObjectNotFoundException;

    /**
     * Add informations on the clustered object with the given name.
     *
     * @param objectName
     *            a name of object
     * @param distributedObjectInfo
     *            informations on the clustered object
     */
    protected abstract void addDistributedObjectInfo(String objectName,
            DistributedObjectInfo distributedObjectInfo);

    /**
     * Sets informations on the clustered object with the given name.
     *
     * @param objectName
     *            a name of object
     * @param distributedObjectInfo
     *            informations on the clustered object
     */
    protected abstract void setDistributedObjectInfo(String objectName,
            DistributedObjectInfo distributedObjectInfo);

    /**
     * Returns a name of interface of this object.
     *
     * @param objectName
     *            a name of object
     * @return a name of interface of this object
     * @throws ObjectNotFoundException
     *             if no object is bound with the given name
     */
    public final String getItfName(final String objectName)
    throws ObjectNotFoundException {
        return getDistributedObjectInfo(objectName).getItfName();
    }

    /**
     * Returns the interface of an object bound with the given name.
     *
     * @param objectName
     *            a name of object
     * @return the interface of an object bound with the given name
     * @throws ObjectNotFoundException
     *             if none object has the given name
     */
    public final Class<?> getInterface(final String objectName)
    throws ObjectNotFoundException {
        return interfaces.get(getItfName(objectName));
    }

    /**
     * Returns a name of business interface of this object (for ejb2 only).
     *
     * @param objectName
     *            a name of object
     * @return a name of business interface of this object
     * @throws ObjectNotFoundException
     *             if no object is bound with the given name
     */
    public final String getBusinessName(final String objectName)
    throws ObjectNotFoundException {
        return getDistributedObjectInfo(objectName).getBusinessName();
    }

    /**
     * Returns the business interface of an object bound with the given name
     * (for ejb2 only).
     *
     * @param objectName
     *            a name of object
     * @return the business interface of an object bound with the given name
     * @throws ObjectNotFoundException
     *             if none object has the given name
     */
    @SuppressWarnings("unchecked")
    public final Class<? extends EJBObject> getRemoteClass(
            final String objectName) throws ObjectNotFoundException {
        return (Class<? extends EJBObject>) interfaces
        .get(getBusinessName(objectName));
    }

    /**
     * Returns the list of Provider URL.
     *
     * @param protocolName
     *            a name of protocol
     * @return providers of the cluster view for the protocol with the given
     *         name
     * @throws ServerClusterViewManagerException
     *             if none provider exists for the protocol with the given name
     */
    public final List<String> getProviderURLs(final String protocolName)
    throws ServerClusterViewManagerException {
        String providerName = CMIConfig.getBindNameForProvider();
        ArrayList<String> providerURLs = new ArrayList<String>();
        try {
            for (CMIReference cmiReference : getCMIReferences(providerName,
                    protocolName)) {
                providerURLs.add(cmiReference.getServerRef().getProviderURL());
            }
        } catch (ObjectNotFoundException e) {
            LOGGER.error(
                    "Cannot get CMIReferences for the provider whith name {0}",
                    providerName, e);
            throw new ServerClusterViewManagerException(
                    "Cannot get CMIReferences for the provider whith name "
                    + providerName, e);
        }
        return providerURLs;
    }

    /**
     * Returns the class that defines the policy for an object with the given
     * name.
     *
     * @param objectName
     *            a name of object
     * @return the class that defines the policy for an object with the given
     *         name
     * @throws ObjectNotFoundException
     *             if none object has the given name
     * @throws ServerClusterViewManagerException
     *             if the class cannot be loaded
     */
    @SuppressWarnings("unchecked")
    public final Class<? extends IPolicy<?>> getPolicyClass(
            final String objectName)

            throws ObjectNotFoundException, ServerClusterViewManagerException {
        return policyStrategyManager.getPolicyClass(objectName);
    }

    /**
     * Returns the class that defines the strategy for an object with the given
     * name.
     *
     * @param objectName
     *            a name of object
     * @return the class that defines the strategy for an object with the given
     *         name
     * @throws ObjectNotFoundException
     *             if none object has the given name
     * @throws ServerClusterViewManagerException
     *             if the class cannot be loaded
     */
    @SuppressWarnings("unchecked")
    public final Class<? extends IStrategy<?>> getStrategyClass(
            final String objectName) throws ObjectNotFoundException,
            ServerClusterViewManagerException {
        return policyStrategyManager.getStrategyClass(objectName);
    }

    /**
     * Returns the date of properties for an object with the given name.
     *
     * @param objectName
     *            a name of object
     * @return the date of properties for an object with the given name
     * @throws ObjectNotFoundException
     *             if none object has the given name
     */
    public final long getDateOfProperties(final String objectName)
    throws ObjectNotFoundException {
        return getDistributedObjectInfo(objectName).getPolicyData()
        .getDateOfProperties();
    }

    /**
     * Returns the properties of the LB policy for an object with the given
     * name.
     *
     * @param objectName
     *            a name of object
     * @return the properties of the LB policy for an object with the given name
     * @throws ObjectNotFoundException
     *             if none object has the given name
     */
    public final Map<String, Object> getPropertiesForPolicy(
            final String objectName) throws ObjectNotFoundException {
        return getDistributedObjectInfo(objectName).getPolicyData()
        .getProperties();
    }

    /**
     * Returns a value property of the LB policy for an object with the given
     * name.
     *
     * @param objectName
     *            a name of object
     * @param propertyName
     *            a name of property
     * @return a value property of the LB policy for an object with the given
     *         name
     * @throws ObjectNotFoundException
     *             if none object has the given name
     */
    public final Object getPropertyForPolicy(final String objectName,
            final String propertyName) throws ObjectNotFoundException {
        return getDistributedObjectInfo(objectName).getPolicyData()
        .getProperties().get(propertyName);
    }

    /**
     * Sets properties for a given object.
     *
     * @param objectName
     *            a name of object
     * @param properties
     *            properties for the LB policy of the given object
     * @throws ObjectNotFoundException
     *             if the specified object doesn't exist
     */
    public final void setPropertiesForPolicy(final String objectName,
            final Map<String, Object> properties)
    throws ObjectNotFoundException {
        DistributedObjectInfo distributedObjectInfo = getDistributedObjectInfo(objectName);
        synchronized (distributedObjectInfo) {
            for (String propertyName : properties.keySet()) {
                distributedObjectInfo.getPolicyData().setProperty(propertyName,
                        properties.get(propertyName));
            }
            setDistributedObjectInfo(objectName, distributedObjectInfo);
        }
    }

    /**
     * Sets a property for a given object.
     *
     * @param objectName
     *            a name of object
     * @param propertyName
     *            a name of property for the LB policy of the given object
     * @param valueName
     *            a value for the given name of property
     * @throws ObjectNotFoundException
     *             if the specified object doesn't exist
     */
    public final void setPropertyForPolicy(final String objectName,
            final String propertyName, final Object valueName)
    throws ObjectNotFoundException {
        DistributedObjectInfo distributedObjectInfo = getDistributedObjectInfo(objectName);
        synchronized (distributedObjectInfo) {
            distributedObjectInfo.getPolicyData().setProperty(propertyName,
                    valueName);
            setDistributedObjectInfo(objectName, distributedObjectInfo);
        }
    }

    /**
     * Returns a name of class that implements the interface
     * InitialContextFactory for the given name of protocol.
     *
     * @param protocolName
     *            a name of protocol
     * @return a name of class that implements the interface
     *         InitialContextFactory for the given name of protocol
     */
    public final String getInitialContextFactoryName(final String protocolName) {
        return initialContextFactories.get(protocolName);
    }

    /**
     * Returns true if the object with the specified name is clustered.
     *
     * @param objectName
     *            a name of object
     * @return true if the object with the specified name is clustered
     */
    public final boolean isClustered(final String objectName) {
        return containObject(objectName);
    }

    /**
     * Returns the name of cluster for the object with the given name.
     *
     * @param objectName
     *            a name of object
     * @return the name of cluster for a object with the given name
     * @throws ObjectNotFoundException
     *             if none object has the given name
     */
    public final String getClusterName(final String objectName)
    throws ObjectNotFoundException {
        return getDistributedObjectInfo(objectName).getClusterName();
    }

    /**
     * Returns the name of class of policy for the object with the given name.
     *
     * @param objectName
     *            a name of object
     * @return the name of class of policy for the object with the given name
     * @throws ObjectNotFoundException
     *             if none object has the given name
     */
    public final String getPolicyClassName(final String objectName)
    throws ObjectNotFoundException {
        return getDistributedObjectInfo(objectName).getPolicyData()
        .getPolicyType();
    }

    /**
     * Sets a new policy for a given object.
     *
     * @param objectName
     *            a name of object
     * @param policyType
     *            a name of class of policy
     * @throws ObjectNotFoundException
     *             if no object is bound with the given name
     */
    public final void setPolicyClassName(final String objectName,
            final String policyType) throws ObjectNotFoundException {
        DistributedObjectInfo distributedObjectInfo = getDistributedObjectInfo(objectName);
        synchronized (distributedObjectInfo) {
            distributedObjectInfo.getPolicyData().setPolicyType(policyType);
            setDistributedObjectInfo(objectName, distributedObjectInfo);
        }
    }

    /**
     * Returns the name of class of strategy for the object with the given name.
     *
     * @param objectName
     *            a name of object
     * @return the name of class of strategy for the object with the given name
     * @throws ObjectNotFoundException
     *             if none object has the given name
     */
    public final String getStrategyClassName(final String objectName)
    throws ObjectNotFoundException {
        return getDistributedObjectInfo(objectName).getPolicyData()
        .getStrategyType();
    }

    /**
     * Sets a new policy for a given object.
     *
     * @param objectName
     *            a name of object
     * @param strategyClassName
     *            a name of class of strategy
     * @throws ObjectNotFoundException
     *             if no object is bound with the given name
     */
    public final void setStrategyClassName(final String objectName,
            final String strategyClassName) throws ObjectNotFoundException {
        DistributedObjectInfo distributedObjectInfo = getDistributedObjectInfo(objectName);
        synchronized (distributedObjectInfo) {
            distributedObjectInfo.getPolicyData().setStrategyType(
                    strategyClassName);
            setDistributedObjectInfo(objectName, distributedObjectInfo);
        }
    }

    /**
     * Sets the algorithm of load-balancing for the object with the given name.
     *
     * @param objectName
     *            a name of object
     * @param policyClassName
     *            a name of class of policy
     * @param strategyClassName
     *            a name of class of strategy
     * @param properties
     *            a set of properties
     * @throws ObjectNotFoundException
     *             if none object has the given name
     */
    public final void setAlgorithmForPolicy(final String objectName,
            final String policyClassName, final String strategyClassName,
            final Map<String, Object> properties)
    throws ObjectNotFoundException {
        DistributedObjectInfo distributedObjectInfo = getDistributedObjectInfo(objectName);
        synchronized (distributedObjectInfo) {
            PolicyData policyData = distributedObjectInfo.getPolicyData();
            policyData.setPolicyType(policyClassName);
            policyData.setStrategyType(strategyClassName);
            policyData.setProperties(properties);
            setDistributedObjectInfo(objectName, distributedObjectInfo);
        }
    }

    /**
     * Returns the configuration of pool of CMIReferenceable for a object with
     * the given name.
     *
     * @param objectName
     *            a name of object
     * @return the configuration of pool of CMIReferenceable for a object with
     *         the given name
     * @throws ObjectNotFoundException
     *             if none object has the given name
     */
    public final IPoolConfiguration getPoolConfiguration(final String objectName)
    throws ObjectNotFoundException {
        return getDistributedObjectInfo(objectName).getPoolConfiguration();
    }

    /**
     * Set the configuration of pool of CMIReferenceable for a object with the
     * given name.
     *
     * @param objectName
     *            a name of object
     * @param poolConfiguration
     *            the configuration of pool of CMIReferenceable for a object
     *            with the given name
     * @throws ObjectNotFoundException
     *             if none object has the given name
     */
    public final void setPoolConfiguration(final String objectName,
            final IPoolConfiguration poolConfiguration)
    throws ObjectNotFoundException {
        DistributedObjectInfo distributedObjectInfo = getDistributedObjectInfo(objectName);
        synchronized (distributedObjectInfo) {
            distributedObjectInfo.setPoolConfiguration(poolConfiguration);
            setDistributedObjectInfo(objectName, distributedObjectInfo);
        }
    }

    /**
     * Returns the address of the local registries.
     *
     * @return the address of the local registries
     */
    public final InetAddress getInetAddress() {
        return inetAddress;
    }

    /**
     * @return the protocols registered in the manager
     */
    public final Set<String> getProtocols() {
        return new HashSet<String>(refsOnLocalRegistries.keySet());
    }

    /**
     * @param protocolName
     *            a name of protocol
     * @return the JMX service URL to access to this MBean or null if the
     *         protocol doesn't exist
     */
    public final JMXServiceURL getJMXServiceURL(final String protocolName) {
        return CMIAdminConnectorManager.getJMXServiceURL(protocolName);
    }

    /**
     * @param protocolName
     *            a name of protocol
     * @return the reference on the local registry for the given protocol
     */
    public final ServerId getRefOnLocalRegistry(final String protocolName) {
        return refsOnLocalRegistries.get(protocolName);
    }

    /**
     * Initialize the statistics.
     */
    protected abstract void initStats();

    /**
     * Return true if the object with the given name is replicated for
     * high-availability. The module 'ha' is so required.
     *
     * @param objectName
     *            a name of object
     * @return true if the object with the given name is replicated for
     *         high-availability
     * @throws ObjectNotFoundException
     *             if none object has the given name
     */
    public final boolean isReplicated(final String objectName)
    throws ObjectNotFoundException {
        return isReplicationManagerStarted()
        && getDistributedObjectInfo(objectName).isReplicated();
    }

    /**
     * Return classnames of the application exceptions.
     *
     * @param objectName
     *            a name of object
     * @return classnames of the application exceptions
     * @throws ObjectNotFoundException
     *             if none object has the given name
     */
    public final Set<String> getApplicationExceptionNames(
            final String objectName) throws ObjectNotFoundException {
        return getDistributedObjectInfo(objectName)
        .getApplicationExceptionNames();
    }

    /**
     * Return true if the object with the given name is stateful.
     *
     * @param objectName
     *            a name of object
     * @return true if the object with the given name is stateful
     * @throws ObjectNotFoundException
     *             if none object has the given name
     */
    public final boolean hasState(final String objectName)
    throws ObjectNotFoundException {
        return getDistributedObjectInfo(objectName).hasState();
    }

    /**
     * @return true if the replication manager is started
     */
    public final boolean isReplicationManagerStarted() {
        return replicationManagerStarted;
    }

    /**
     * Set if the replication manager is started.
     *
     * @param replicationManagerStarted
     *            true if the replication manager is started
     */
    public final void setReplicationManagerStarted(
            final boolean replicationManagerStarted) {
        this.replicationManagerStarted = replicationManagerStarted;
    }

    /**
     * Gets available load balancing policies and strategies.
     *
     * @return Available policies and strategies.
     */
    public final Map<String, Set<String>> getAvailablePoliciesAndStrategies() {
        // TODO Load user defined policies and strategies.
        Map<String, Set<String>> ret = new HashMap<String, Set<String>>();
        ret.put("policies", new HashSet<String>(policyStrategyManager.getPolicyClassNames()));
        ret.put("strategies", new HashSet<String>(policyStrategyManager.getStrategyClassNames()));
        return ret;
    }

    /**
     * Load classes embedded in the CMI distribution.
     */
    @SuppressWarnings("unchecked")
    private void loadEmbeddedLBClasses() {
        for (Class<?> embeddedPolicyClass : embeddedPolicyClasses) {
            policyStrategyManager.addPolicyClass((Class<? extends IPolicy<CMIReference>>) embeddedPolicyClass, true);
        }
        for (Class<?> embeddedStrategyClass : embeddedStrategyClasses) {
            policyStrategyManager.addStrategyClass( (Class<? extends IStrategy<CMIReference>>) embeddedStrategyClass, true);
        }
    }

    /**
     * Return true if the given class name is embedded policy class.
     *
     * @param className
     *            a name of class
     * @return true if the given class name is embedded policy class
     */
    public final boolean isEmbeddedPolicy(final String className) {
        for (Class<?> embeddedPolicyClass : embeddedPolicyClasses) {
            if (embeddedPolicyClass.getName().equals(className)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Return true if the given class name is embedded strategy class.
     *
     * @param className
     *            a name of class
     * @return true if the given class name is embedded strategy class
     */
    public final boolean isEmbeddedStrategy(final String className) {
        for (Class<?> embeddedStrategyClass : embeddedStrategyClasses) {
            if (embeddedStrategyClass.getName().equals(className)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Adds an archive containing load balancing classes. This archive might
     * have been sent by a remote client or server.
     *
     * @param bytesOfFile
     *            Bytes content of the file
     * @param location
     *            File location
     * @param strategies
     *            Load balancing policies
     * @param policies
     *            Load balancing strategies
     * @return The archive Id used by the <code>PolicyStrategyManager</code>
     *         to register the new archive.
     *
     */
    public Object addLoadBalancingArchive(final byte[] bytesOfFile,
            final String location, final Class<?>[] policies,
            final Class<?>[] strategies) {
        LOGGER
        .debug(
                "Sending new load balancing archive located at '{0}'. Step 3: server-side manager is delegating reception to policy/stretegy manager ",
                location);
        return policyStrategyManager.addLoadBalancingArchive(bytesOfFile,
                location, policies, strategies);
    }

    /**
     * Remove Load Balancing Archive, previously added.
     *
     * @param archiveId
     *            The Id of the archive.
     * @param policies
     *            Names of the policies to remove
     * @param strategies
     *            Names of the strategies to remove
     * @param isArchiveOwner
     *            if true then the call is performed by the archive owner. The
     *            archive owner is the first server that deployed the archive.
     * @throws Exception
     *             any.
     *
     */
    public void removeLoadBalancingArchive(final Object archiveId,
            final String[] policies, final String[] strategies,
            final boolean isArchiveOwner) throws Exception {
        policyStrategyManager.removeLoadBalancingArchive(archiveId, policies,
                strategies, isArchiveOwner);
    }

    /**
     * Share the archive to all the cluster.
     *
     * @param archiveId
     *            the archive Id
     * @param bytesOfFile
     *            bytes content of the archive.
     */
    public abstract void broadCastArchive(final Object archiveId,
            final byte[] bytesOfFile);

    /**
     * Checks if the given archive is already distributed
     *
     * @param archiveId
     *            the Id of the archive to check.
     * @return true if the archive is already distributed, false otherwise
     */
    public abstract boolean isAlreadyDistributed(final Object archiveId);

    /**
     * remove distributed archive.
     *
     * @param archiveId
     *            The archive Id
     */
    public abstract void removeDistributedArchive(final Object archiveId);

    /**
     * @return the policyStrategyManager
     */
    public IPolicyStrategyManager getPolicyStrategyManager() {
        return policyStrategyManager;
    }

    /**
     * @param policyStrategyManager
     *            the policyStrategyManager to set
     */
    public void setPolicyStrategyManager(
            final IPolicyStrategyManager policyStrategyManager) {
        this.policyStrategyManager = policyStrategyManager;
    }

    /**
     * @return the policyStrategyLoader
     */
    public IPolicyStrategyLoader getPolicyStrategyLoader() {
        return policyStrategyLoader;
    }

    /**
     * @param policyStrategyLoader
     *            the policyStrategyLoader to set
     */
    public void setPolicyStrategyLoader(
            final IPolicyStrategyLoader policyStrategyLoader) {
        this.policyStrategyLoader = policyStrategyLoader;
    }

    /**
     * Return the list of servers for a given protocol.
     *
     * @param protocolName
     *            a name of protocol
     * @return the set of references on server connected to this server
     */
    public final Set<ServerId> getServerRefs(final String protocolName) {
        Set<ServerId> serverRefs = new HashSet<ServerId>();
        for (String objectName : getObjectNames()) {
            try {
                for (CMIReference cmiRef : getCMIReferences(objectName,
                        protocolName, false)) {
                    serverRefs.add((ServerId) cmiRef.getServerRef());
                }
            } catch (ObjectNotFoundException e) {
                // Not an error!
            }
        }
        return serverRefs;
    }

    /**
     * @param serverName
     *            a server name
     * @return references on registries of the given server
     */
    public final Set<ServerId> getRefsOnRegistries(final String serverName) {
        Set<ServerId> serverRefs = new HashSet<ServerId>();
        for (String objectName : getObjectNames()) {
            try {
                for (CMIReference cmiRef : getCMIReferences(objectName)) {
                    ServerId serverId = (ServerId) cmiRef.getServerRef();
                    if (serverId.getServerName().equals(serverName)) {
                        serverRefs.add(serverId);
                    }
                }
            } catch (ObjectNotFoundException e) {
                // Not an error!
            }
        }
        return serverRefs;
    }

    /**
     * Gets all objects hosted on a given server started with the given
     * protocol.
     *
     * @param serverUrl
     *            a server reference
     * @param protocolName
     *            a given protocol
     * @return the set of references on objects for the given server.
     */
    public final Set<String> getServerObjectForProtocol(final String serverUrl,
            final String protocolName) {
        Set<String> objectRefs = new HashSet<String>();
        for (String objectName : getObjectNames()) {
            try {
                for (CMIReference cmiRef : getCMIReferences(objectName,
                        protocolName, false)) {
                    if (cmiRef.getServerRef().getProviderURL()
                            .equals(serverUrl)) {
                        objectRefs.add(objectName);
                    }
                }
            } catch (ObjectNotFoundException e) {
                // Not an error!
            }
        }
        return objectRefs;
    }

    /**
     * Returns a list of CMIReference for an object with the given name and
     * protocol. The objects deployed on a blacklisted server are discarded.
     *
     * @param objectName
     *            a name of object
     * @param protocolName
     *            a name of protocol
     * @return a list of CMIReference for an object with the given name and
     *         protocol
     * @throws ObjectNotFoundException
     *             if none object has the given name for the given protocol
     */
    public final List<CMIReference> getCMIReferences(final String objectName,
            final String protocolName) throws ObjectNotFoundException {
        return getCMIReferences(objectName, protocolName, true);
    }

    /**
     * Returns a list of CMIReference for an object with the given name and
     * protocol.
     *
     * @param objectName
     *            a name of object
     * @param protocolName
     *            a name of protocol
     * @param filter
     *            true if the objects deployed on a blacklisted server are
     *            discarded
     * @return a list of CMIReference for an object with the given name and
     *         protocol
     * @throws ObjectNotFoundException
     *             if none object has the given name for the given protocol
     */
    protected abstract List<CMIReference> getCMIReferences(
            final String objectName, final String protocolName,
            final boolean filter) throws ObjectNotFoundException;

    /**
     * @return domain names
     */
    public final Set<String> getDomainNames() {
        Set<String> domainNames = new HashSet<String>();
        for (String objectName : getObjectNames()) {
            try {
                for (CMIReference cmiRef : getCMIReferences(objectName)) {
                    domainNames.add(((ServerId) cmiRef.getServerRef())
                            .getDomainName());
                }
            } catch (ObjectNotFoundException e) {
                // Not an error!
            }
        }
        return domainNames;
    }

    /**
     * @param domainName
     *            domain name
     * @return names of CMI server
     */
    public final Set<String> getServerNames(final String domainName) {
        Set<String> serverNames = new HashSet<String>();
        for (String objectName : getObjectNames()) {
            try {
                for (CMIReference cmiRef : getCMIReferences(objectName)) {
                    ServerId serverId = (ServerId) cmiRef.getServerRef();
                    if (serverId.getDomainName().equals(domainName)) {
                        serverNames.add(serverId.getServerName());
                    }
                }
            } catch (ObjectNotFoundException e) {
                // Not an error!
            }
        }
        return serverNames;
    }

    /**
     * @return references of blacklisted server
     */
    public final Set<ServerId> getBlacklistedServers() {
        Set<ServerId> serverIds = new HashSet<ServerId>();
        for (String objectName : getObjectNames()) {
            try {
                for (CMIReference cmiRef : getCMIReferences(objectName)) {
                    ServerId serverId = (ServerId) cmiRef.getServerRef();
                    if (isServerBlackListed(serverId)) {
                        serverIds.add(serverId);
                    }
                }
            } catch (ObjectNotFoundException e) {
                // Not an error!
            }
        }
        return serverIds;
    }

    /**
     * @param serverRef
     *            a reference on CMI server
     * @return references on clustered object for a given server
     */
    public final Set<CMIReference> getCMIRefsForServerRef(
            final ServerRef serverRef) {
        Set<CMIReference> cmiRefs = new HashSet<CMIReference>();
        for (String objectName : getObjectNames()) {
            try {
                for (CMIReference cmiRef : getCMIReferences(objectName)) {
                    ServerRef serverId = cmiRef.getServerRef();
                    if (serverId.equals(serverRef)) {
                        cmiRefs.add(cmiRef);
                    }
                }
            } catch (ObjectNotFoundException e) {
                // Not an error!
            }
        }
        return cmiRefs;
    }

}
